#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Feb 17 11:06:49 2022

@author: felix
"""
import numpy as np

class AbstractMaterial:
    
    def computeStress(self, strain):
        assert 'stress needs to be implemented'    

class LinearIsotropicConductivityMaterial(AbstractMaterial):
    
    def __init__(self, kappa):
        self.kappa = kappa
        self.eigenvalues = [kappa]
        
    def computeStress(self, strain):
        stress = np.zeros(strain.shape)
        for i in range(3):
            stress[i] = self.kappa * strain[i]
        return stress
    
    def computeStrain(self, stress):
        strain = np.zeros(stress.shape)
        for i in range(3):
            strain[i] = 1./self.kappa * stress[i]
        return strain

class LinearElasticIsotropicMaterial(AbstractMaterial):
    
    def __init__(self, E, nu):
        self.E = E
        self.nu = nu
        self.la = E*nu/((1-2*nu)*(1+nu))
        self.mu = 0.5*E/(1+nu)
        self.K = E/(3*(1-2*nu))
        self.eigenvalues = [2*self.mu, 3*self.K]
    
    def computeStress(self, strain):
        stress = np.zeros(strain.shape)
        stran_trace = strain[0] + strain[1] + strain[2]
        for i in range(3):
            stress[i] = 2*self.mu * strain[i] + self.la * stran_trace
            stress[i+3] = self.mu * strain[i+3]
        return stress
    
    def computeStrain(self, stress):
        strain = np.zeros(stress.shape)
        stress_trace = stress[0] + stress[1] + stress[2]
        for i in range(3):
            strain[i]   = 1/(2*self.mu)*(stress[i] - self.la/(3*self.la + 2*self.mu)*stress_trace)
            strain[i+3] = 1/self.mu *stress[i+3]
        return strain

class ReferenceMaterial(LinearElasticIsotropicMaterial):

    def __init__(self,materialDict):
        self.materialDict = materialDict
        self.initialize()
        
    def initialize(self):
        eigenvalues = []
        for mat_key in self.materialDict.keys():
            eigenvalues += self.materialDict[mat_key].eigenvalues

        maxEV = max(eigenvalues)
        minEV = min(eigenvalues)
        self.mu = 0.25*(maxEV+minEV)
        self.la = 0.
        self.alpha = 2*self.mu
        
    def update(self, alpha0):
        self.mu = 0.5*alpha0
        self.alpha = alpha0
    
    def applyReferenceMaterial(self, field):
        return self.computeStress(field)
    
    def applyInverseReferenceMaterial(self, field):
        return self.computeStrain(field)
    
    def computeStiffnessNorm(self, field):
        return np.sqrt(np.dot(self.applyReferenceMaterial(field).ravel(),field.ravel())/(np.prod(field.shape)))
    
    def computeComplianceNorm(self, field):
        return np.sqrt(np.dot(self.applyInverseReferenceMaterial(field).ravel(),field.ravel())/(np.prod(field.shape) ) )          

    def l2Product(self, field1, field2):
        assert field1.shape == field2.shape, "non-matching shapes"
        return np.dot(field1.ravel(),field2.ravel())/np.prod(field1.shape)
    
    def bb_Step(self,stress, oldStress, strain, oldStrain):
        a = self.l2Product(stress-oldStress, strain-oldStrain)
        # Dirty workaround for the voigt notation
        self.update(1.)
        b = self.computeStiffnessNorm(strain-oldStrain)**2
        self.update(a/b)
        
class ConductivityReferenceMaterial(LinearIsotropicConductivityMaterial):
    
    def __init__(self,materialDict):
        self.materialDict = materialDict
        self.initialize()
        
    def initialize(self):
        eigenvalues = []
        for mat_key in self.materialDict.keys():
            eigenvalues += self.materialDict[mat_key].eigenvalues

        maxEV = max(eigenvalues)
        minEV = min(eigenvalues)
        self.kappa = 0.5*(maxEV+minEV)
        
    def update(self, alpha0):
        self.kappa = alpha0
        
    def computeStiffnessNorm(self, field):
        return np.sqrt(self.kappa * np.dot(field.ravel(),field.ravel())/(np.prod(field.shape)))
    
    def computeComplianceNorm(self, field):
        return np.sqrt(1/self.kappa * np.dot(field.ravel(),field.ravel())/(np.prod(field.shape)))
    
    def applyReferenceMaterial(self, field):
        return self.computeStress(field)
    
    def l2Product(self, field1, field2):
        assert field1.shape == field2.shape, "non-matching shapes"
        return np.dot(field1.ravel(), field2.ravel())/np.prod(field1.shape)
    
    def bb_Step(self,stress,oldStress,strain,oldStrain):
        self.update( self.l2Product(stress-oldStress,strain-oldStrain)/self.l2Product(strain-oldStrain,strain-oldStrain))

class ResidualController:
    
    def __init__(self, colors, referenceMaterial):
        self.referenceMaterial = referenceMaterial
        self.N_factor = np.prod(colors.shape)
        self.currentStressnorm = 0.

    def updateStressnorm(self, stressField):
        stressMean = self.computeMean(stressField)
        #print('stressMean',stressMean)
        self.currentStressnorm = self.referenceMaterial.computeComplianceNorm(stressMean)
        
    def computeMean(self,field):
        N = field.shape[0]
        meanField = np.zeros(N)
        for i in range(N):
            meanField[i] = np.sum(field[i])/self.N_factor
        return meanField

class AbstractMaterialApplication:
    
    def __init__(self ,colors, materialDict, referenceMaterial):
        self.colors = colors
        self.materials = materialDict
        self.referenceMaterial = referenceMaterial
        self.residualController = ResidualController(colors,referenceMaterial)
        self.returnStress = False

    def computeStress(self,field):
        assert 'stress computation needs to be implemented'
    
    def computePolarization(self, field):
        stress = self.computeStress(field)
        self.residualController.updateStressnorm(stress)
        stress0 = self.referenceMaterial.applyReferenceMaterial(field.real_field)
        field.real_field[:] = stress - stress0
        
        if self.returnStress:
            return stress
        

class MaterialApplication_Slow(AbstractMaterialApplication):
    # This implementation would be quite fast if written in Cython or C
    def computeStress(self,field):
        Nx,Ny,Nz = self.colors.shape
        stress = np.zeros(field.real_field.shape)
        for i in range(Nx):
            for j in range(Ny):
                for k in range(Nz):
                    stress[:,i,j,k] = self.materials[str(self.colors[i,j,k])].computeStress(field.real_field[:,i,j,k])
        return stress

class MaterialApplicationFast(AbstractMaterialApplication):
    def computeStress(self,field):
        stress = np.zeros(field.real_field.shape)
        for key in self.materials.keys():
            
            stress += np.where(self.colors==int(key), self.materials[key].computeStress(field.real_field), 0) 
        return stress
