#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Feb 17 09:04:49 2022

@author: felix
"""

import numpy as np

class AbstractNLO:
    def __init__(self, colors, referenceMaterial):
        self.Nx, self.Ny, self.Nz = np.shape(colors)

        self.xfreq = np.fft.fftfreq(self.Nx)
        self.yfreq = np.fft.fftfreq(self.Ny)
        self.zfreq = np.fft.rfftfreq(self.Nz)
        
        self.referenceMaterial = referenceMaterial
    
    def applyNLO(self,primaryField):
        assert 'applyNLO needs to be implemented'

class MoulinecSuquetNLO(AbstractNLO):
    def applyNLO(self, primaryField):
        
        f = np.zeros(3, dtype = np.complex128)
        G = np.zeros([3,3], dtype  =  np.complex128)
        u = np.zeros(3, dtype = np.complex128)
        
        tau = np.zeros([3,3], dtype = np.complex128)
        
        for i in range(len(self.xfreq)):
            for j in range(len(self.yfreq)):
                for k in range(len(self.zfreq)):
                    
                    if i==0 and j == 0 and k==0: 
                        pass
                    
                    elif (self.Nx%2==0 and i==self.Nx//2) or (self.Ny%2==0 and j==self.Ny//2) or (self.Nz%2==0 and k==self.Nz//2):
                        primaryField.cpx_field[:,i,j,k] = 0
                        
                    else:
                        xi = np.array([self.xfreq[i],self.yfreq[j],self.zfreq[k]])
                        
                        normXiSquared = np.dot(xi,xi)
                        
                        tau_voigt = primaryField.cpx_field[:,i,j,k]
                        tau[0][0] = tau_voigt[0]
                        tau[1][1] = tau_voigt[1]
                        tau[2][2] = tau_voigt[2]
                        
                        tau[0][1] = tau_voigt[5]
                        tau[1][0] = tau_voigt[5]
                        
                        tau[0][2] = tau_voigt[4]
                        tau[2][0] = tau_voigt[4]
                        
                        tau[2][1] = tau_voigt[3]
                        tau[1][2] = tau_voigt[3]
                        
                        f = np.dot(tau,xi)
                        
                        G = - np.eye(3)/(normXiSquared*self.referenceMaterial.mu) 
                        G += (self.referenceMaterial.mu + self.referenceMaterial.la)/(self.referenceMaterial.mu*(2*self.referenceMaterial.mu+self.referenceMaterial.la))*np.outer(xi,xi)/normXiSquared**2
                        
                        u = np.dot(G,f)
                        
                        tau_voigt[0] = u[0]*xi[0]
                        tau_voigt[1] = u[1]*xi[1]
                        tau_voigt[2] = u[2]*xi[2]
                        
                        tau_voigt[3] = u[2]*xi[1] + u[1]*xi[2]
                        tau_voigt[4] = u[2]*xi[0] + u[0]*xi[2]
                        tau_voigt[5] = u[0]*xi[1] + u[1]*xi[0]


from numba import jit,njit
from NLO.NLO_cython import applyMoulinecSuquetNLO, applyRotatedStaggeredGridNLO, applyStaggeredGridNLO

class MoulinecSuquetCythonNLO(AbstractNLO):
    def applyNLO(self, primaryField):
        applyMoulinecSuquetNLO(np.array([self.Nx,self.Ny,self.Nz]),
                               primaryField.cpx_field,
                               self.xfreq,self.yfreq,self.zfreq,
                               self.referenceMaterial.mu)
    

class RotatedStaggeredGridNLO(AbstractNLO):
    def __init__(self, colors, referenceMaterial):
        super(RotatedStaggeredGridNLO, self).__init__(colors, referenceMaterial)
        
        self.exp1 = np.exp(2j*np.pi*self.xfreq)
        self.exp2 = np.exp(2j*np.pi*self.yfreq)
        self.exp3 = np.exp(2j*np.pi*self.zfreq)

        
    def applyNLO(self, primaryField):
        
        applyRotatedStaggeredGridNLO(np.array([self.Nx,self.Ny,self.Nz]),
                       primaryField.cpx_field,
                       self.exp1,self.exp2,self.exp3,
                       self.referenceMaterial.mu)

class StaggeredGridNLO(AbstractNLO):
    def __init__(self,colors,referenceMaterial):
        super(StaggeredGridNLO,self).__init__(colors,referenceMaterial)
        self.k1 = np.exp(-2j*np.pi*self.xfreq)-1.
        self.k2 = np.exp(-2j*np.pi*self.yfreq)-1.
        self.k3 = np.exp(-2j*np.pi*self.zfreq)-1.
    def applyNLO(self,primaryField):
        applyStaggeredGridNLO(np.array([self.Nx,self.Ny,self.Nz]),
                       primaryField.cpx_field,
                       self.k1,self.k2,self.k3,
                       self.referenceMaterial.mu)
        
        
#################################################
# Heat Conductivity 
#################################################
from NLO.NLO_cython import applyMoulinecSuquetNLO_HEAT
 
class HeatNLO_MoulinecSuquet(AbstractNLO):
    def applyNLO(self,primaryField):
        applyMoulinecSuquetNLO_HEAT(np.array([self.Nx,self.Ny,self.Nz]),
                               primaryField.cpx_field,
                               self.xfreq,self.yfreq,self.zfreq,
                               self.referenceMaterial.kappa)