#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Feb 17 14:56:22 2022

@author: felix
"""
import time
import numpy as np

class Timer:
    def __init__(self):
        self.elapsedTime = 0.
        
    def start(self):
        self._start_time = time.perf_counter()
        
    def stop(self):
        stopTime = time.perf_counter()
        self.elapsedTime += stopTime - self._start_time
    
class MyTimer:
    def __init__(self):
        self.nloTimer = Timer()
        self.fftTimer = Timer()
        self.polarizationTimer = Timer()
        self.residualTimer = Timer()
        self.solverTimer =  Timer()
        
    def evaluate(self):
        print(f'Total NLO time  {self.nloTimer.elapsedTime} s')
        print(f'Total polarization time   {self.polarizationTimer.elapsedTime} s')
        print(f'Total fft time  {self.fftTimer.elapsedTime} s')
        print(f'Total residualTimer time  {self.residualTimer.elapsedTime} s')
        print(f'Total solver time  {self.solverTimer.elapsedTime} s')

class AbstractSolver:
    def __init__(self,
                 primaryField,
                 materialApplication,
                 nlo, 
                 boundaryCondition,
                 maxit=100,
                 tol=1e-5):
        
        self.primaryField = primaryField
        self.materialApplication = materialApplication
        self.nlo = nlo
        self.maxit = maxit
        self.boundaryCondition = boundaryCondition
        self.tol = tol
        self.timer = MyTimer()
        
        self.res = []
        
        
    def solve(self):
        assert 'solve routine must be implemented'
        
class BasicScheme(AbstractSolver):
    
    def solve(self):
        iter = 0 # current number of iterations
        res = 1. # residual
        self.res.append(res)
        self.timer.solverTimer.start()
        oldField = np.empty_like(self.primaryField.real_field)
        
        while res > self.tol and iter <= self.maxit:
    
            iter += 1
        
            #########################
            # Polarization
            #########################
            oldField[:]= self.primaryField.real_field
            
            self.timer.polarizationTimer.start()
            #print(f'IIITER {iter}')
            self.materialApplication.computePolarization(self.primaryField)
            #print('Mean Polarization',self.materialApplication.residualController.computeMean(self.primaryField.real_field))
            self.timer.polarizationTimer.stop()
            
            #########################
            # FFT
            #########################
            self.timer.fftTimer.start()
            self.primaryField.fft()
            self.timer.fftTimer.stop()
            #########################
            # Apply Gamma^0
            #########################
            self.timer.nloTimer.start()
            self.nlo.applyNLO(self.primaryField)
            self.timer.nloTimer.stop()
            #########################
            # Set boundary Condition
            ########################
            self.boundaryCondition.handleBC(self.primaryField)
            #########################
            # IFFT
            #########################
            self.timer.fftTimer.start()
            self.primaryField.ifft()
            self.timer.fftTimer.stop()
            #########################
            # Compute residual
            #########################
            self.timer.residualTimer.start()
            res = self.materialApplication.referenceMaterial.computeStiffnessNorm(self.primaryField.real_field-oldField)/self.materialApplication.residualController.currentStressnorm
            self.timer.residualTimer.stop()
            self.res.append(res)
            #########################
            # Print current progress
            #########################
            print(f"iteration: {iter:5} res: {res}")
        self.timer.solverTimer.stop()
        self.timer.evaluate()
        if iter>=self.maxit:
            print(f"Maximum number of iterations of {self.maxit} is reached. The solution may not be reliable")
        
class BB(AbstractSolver):
    def __init__(self,primaryField,
                 materialApplication,
                 nlo, 
                 boundaryCondition,
                 maxit=100,tol=1e-5):
        
        super(BB,self).__init__(primaryField,
                 materialApplication,
                 nlo, 
                 boundaryCondition,
                 maxit,tol)
        self.materialApplication.returnStress = True
        
    def solve(self):
        iter = 0 # current number of iterations
        res = 1. # residual
        self.res.append(res)
        self.timer.solverTimer.start()

        oldField = np.zeros(self.primaryField.real_field.shape)
        oldStress = np.zeros(self.primaryField.real_field.shape)
        stress = np.zeros(self.primaryField.real_field.shape)
        while res > self.tol and iter <= self.maxit:
    
            iter += 1
        
            #########################
            # Polarization
            #########################
            self.timer.polarizationTimer.start()
            if iter>1:
                oldStress[:] = stress
                stress = self.materialApplication.computeStress(self.primaryField)
                self.materialApplication.referenceMaterial.bb_Step(stress,oldStress,self.primaryField.real_field,oldField)

                
            oldField[:]= self.primaryField.real_field
            
            
            stress = self.materialApplication.computePolarization(self.primaryField)

            self.timer.polarizationTimer.stop()
            
            #########################
            # FFT
            #########################
            self.timer.fftTimer.start()
            self.primaryField.fft()
            self.timer.fftTimer.stop()
            #########################
            # Apply Gamma^0
            #########################
            self.timer.nloTimer.start()
            self.nlo.applyNLO(self.primaryField)
            self.timer.nloTimer.stop()
            #########################
            # Set boundary Condition
            ########################
            self.boundaryCondition.handleBC(self.primaryField)
            #########################
            # IFFT
            #########################
            self.timer.fftTimer.start()
            self.primaryField.ifft()
            self.timer.fftTimer.stop()
            #########################
            # Compute residual
            #########################
            self.timer.residualTimer.start()
            res = self.materialApplication.referenceMaterial.computeStiffnessNorm(self.primaryField.real_field-oldField)/self.materialApplication.residualController.currentStressnorm
            self.timer.residualTimer.stop()
            #########################
            self.res.append(res)
            # Print current progress
            #########################
            print(f"iteration: {iter:5} res: {res}")
        self.timer.solverTimer.stop()
        self.timer.evaluate()
        if iter>=self.maxit:
            print(f"Maximum number of iterations of {self.maxit} is reached. The solution may not be reliable")
            
        
from Field.Field import Field
class CG(AbstractSolver):
    def applyNLO(self,field):
        
        
        self.timer.fftTimer.start()
        field.fft()
        self.timer.fftTimer.stop()
        
        self.timer.nloTimer.start()
        self.nlo.applyNLO(field)
        self.timer.nloTimer.stop()
        
        field.cpx_field[:,0,0,0] = 0.
        
        self.timer.fftTimer.start()
        field.ifft()
        self.timer.fftTimer.stop()
    
    def solve(self):
        self.timer.solverTimer.start()
        self.res = []
        self.refMat = self.materialApplication.referenceMaterial
        self.refMat.update(1.)
        
        G = Field(self.materialApplication.colors,self.primaryField.ndim)
        D = Field(self.materialApplication.colors,self.primaryField.ndim)
        Z = Field(self.materialApplication.colors,self.primaryField.ndim)
        
        self.timer.polarizationTimer.start()
        stress = self.materialApplication.computeStress(self.primaryField)
        self.timer.polarizationTimer.stop()
        G.real_field[:] = stress
        sigBar = self.materialApplication.residualController.computeMean(stress)
        self.applyNLO(G)
        
        D.real_field[:] = -G.real_field
        
        r = self.refMat.computeStiffnessNorm(G.real_field)
        res = r/self.refMat.computeStiffnessNorm(sigBar)
        self.res.append(res)
        iter = 0
        while iter<=self.maxit and res>self.tol:
            iter+=1
            r_old = r
            self.timer.polarizationTimer.start()
            stress = self.materialApplication.computeStress(D)
            self.timer.polarizationTimer.stop()
            deltasigBar =self.materialApplication.residualController.computeMean(stress)
            
            Z.real_field[:] = stress
            
            self.applyNLO(Z)
            
            alpha = r**2/self.refMat.l2Product(self.refMat.applyReferenceMaterial(Z.real_field),D.real_field)
            self.primaryField.real_field += alpha*D.real_field
            
            sigBar += alpha*deltasigBar
            
            res = r/self.refMat.computeComplianceNorm(sigBar)
            self.res.append(res)
            G.real_field += alpha*Z.real_field
            r = self.refMat.computeStiffnessNorm(G.real_field)
            
            gamma = r**2/r_old**2
            
            D.real_field *= gamma
            D.real_field += -G.real_field 
            
            print(f"iteration: {iter:5} res: {res}")
        
        self.timer.solverTimer.stop()
        self.timer.evaluate()
        if iter>=self.maxit:
            print(f"Maximum number of iterations of {self.maxit} is reached. The solution may not be reliable")
        
    
        