#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Feb 17 14:09:51 2022

@author: felix
"""

#####################################################################################################################
#                                                                                                                   #
#      FFT-Workshop-22 : Introduction to FFT-based numerical methods for the homogenization of random materials     #                  
#                                                                                                                   #
#                          sample solution - compilation of NLO required before use                                 #
#                                                                                                                   #
#####################################################################################################################

import numpy as np
import os

from Image.MicrostructureImage import MicrostructureImage
from Image.WriteOutput import PyEVTKOutput
from Field.Field import Field

from NLO.NLO import MoulinecSuquetNLO, MoulinecSuquetCythonNLO, RotatedStaggeredGridNLO, StaggeredGridNLO,HeatNLO_MoulinecSuquet
from Material.smallStrainMaterials import LinearElasticIsotropicMaterial, ReferenceMaterial, MaterialApplicationFast, LinearIsotropicConductivityMaterial, ConductivityReferenceMaterial
from Solvers.Solvers import BasicScheme, BB, CG


solversDict = {'BasicScheme': BasicScheme,
               'BB': BB,
               'CG': CG}
                
elasticityDiscretizationDict = {'MoulinecSuquet': MoulinecSuquetCythonNLO,
                              'MoulinecSuquetSlow': MoulinecSuquetNLO,
                      'RotatedStaggered': RotatedStaggeredGridNLO,
                      'Staggered': StaggeredGridNLO}

heatDiscretizationDict = {'MoulinecSuquet': HeatNLO_MoulinecSuquet}


class BoundaryCondition:
    
    def __init__(self, colors, strain0):
        self.strain0 = strain0
        self.normalizationFactor = np.prod(colors.shape)
        
    def handleBC(self, field):
        field.cpx_field[:,0,0,0] = self.strain0 * self.normalizationFactor

class AbstractOutput:
    
    def __init__(self,folder):
        self.folder = folder
        if not os.path.isdir(self.folder):
            os.mkdir(self.folder)
            
        self.timesFile = os.path.join(self.folder,'computationTimes.txt')
        self.residualsFile = os.path.join(self.folder,'residual.txt')                
        self.fieldOutput = PyEVTKOutput(self.folder)  
        self.resultsFile = os.path.join(self.folder,'results.txt')
        
    def writeTimesAndResidual(self,solver):
        res = np.array(solver.res)
        iters = np.arange(len(res))
        np.savetxt(self.residualsFile,np.c_[iters,res],delimiter = '    ', header='iter    residual')
        
        textFile = open(self.timesFile,'w')
        
        textFile.write(f'Total NLO time             {solver.timer.nloTimer.elapsedTime} s\n')
        textFile.write(f'Total polarization time    {solver.timer.polarizationTimer.elapsedTime} s\n')
        textFile.write(f'Total fft time             {solver.timer.fftTimer.elapsedTime} s\n')
        textFile.write(f'Total residualTimer time   {solver.timer.residualTimer.elapsedTime} s\n')
        textFile.write('-----------------------------------------------------------------------\n')
        textFile.write(f'Total solver time          {solver.timer.solverTimer.elapsedTime} s\n')
        textFile.close
        
    def writeFields(self,stressField,strain_field):
        assert 'Needs to be implemented in subclass'
        
    def writeResults(self,stressMean,strainMean):
        assert 'Needs to be implemented in subclass'

class ElasticityOutput(AbstractOutput):
    
    def writeFields(self,stressField,strainField):
        self.fieldOutput.addStrain(strainField)
        self.fieldOutput.addStress(stressField)
        self.fieldOutput.write()
    
    def writeResults(self,stressMean,strainMean):
        np.savetxt(self.resultsFile,np.c_[strainMean,stressMean],delimiter  = '     ', header = 'strain    stress')

class ConductivityOutput(AbstractOutput):
    
    def writeFields(self,stressField,strainField,):
        self.fieldOutput.addGradient(strainField)
        self.fieldOutput.addFlux(stressField)
        self.fieldOutput.write()
    
    def writeResults(self,stressMean,strainMean):
        np.savetxt(self.resultsFile,np.c_[strainMean,stressMean],delimiter  = '     ', header = 'tempgrad    flux')
        
        

def solveProblem(mhaFile, materialDict, referenceMaterial, load, solverClass, discretizationClass, output):
    
    microstructure = MicrostructureImage()
    microstructure.readColors(mhaFile)    
    
    materialApplication = MaterialApplicationFast(microstructure.colors, materialDict, referenceMaterial)
    boundaryCondition = BoundaryCondition(microstructure.colors, load)
    primaryField = Field(microstructure.colors,load.size)
    primaryField.initialize(load)
    
    my_nlo = discretizationClass(microstructure.colors,referenceMaterial)
    
    my_solver = solverClass(primaryField, materialApplication, my_nlo, boundaryCondition, maxit = 300,tol = 1e-6)
    
    my_solver.solve()
    
    stress_field = materialApplication.computeStress(primaryField)
    
    stress_mean = materialApplication.residualController.computeMean(stress_field)
    strain_mean = materialApplication.residualController.computeMean(primaryField.real_field)
    
    output.writeTimesAndResidual(my_solver)
    
    output.writeFields(stress_field,primaryField.real_field)#
    output.writeResults(stress_mean,strain_mean)


if __name__ == "__main__":
    if not os.path.isdir('./Results'):
        os.mkdir('./Results')
        
    ########################
    # Elasticity Problem
    ########################
    mhaFile = "./Microstructures/ball_32.mha"
    materialDict = {'0': LinearElasticIsotropicMaterial(2100, 0.3),
                    '1':LinearElasticIsotropicMaterial(72000, 0.22)}
    load = np.array([0.01,0,0,0.0,0,0])
    
    solverClass = solversDict['CG']
    discretizationClass = elasticityDiscretizationDict['Staggered']
    output = ElasticityOutput('./Results/Elasticity')
    referenceMaterial = ReferenceMaterial(materialDict)
    solveProblem(mhaFile, materialDict, referenceMaterial, load, solverClass, discretizationClass, output)
    
    
    ########################
    #Conductivity Problem
    #######################
    mhaFile = "./Microstructures/ball_32.mha"
    materialDict = {'0': LinearIsotropicConductivityMaterial(1),
                    '1':LinearIsotropicConductivityMaterial(2)}
    load = np.array([0.01,0,0])
    
    solverClass = solversDict['CG']
    discretizationClass = heatDiscretizationDict['MoulinecSuquet']
    output = ConductivityOutput('./Results/Conductivity')
    referenceMaterial = ConductivityReferenceMaterial(materialDict)
    solveProblem(mhaFile, materialDict, referenceMaterial, load, solverClass, discretizationClass, output)




